/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.opengl.entities.base;

import com.jogamp.newt.event.InputEvent;
import com.jogamp.opengl.GL3bc;
import java.util.ArrayList;
import java.util.List;
import fr.amap.lidar.amapvox.visualization.opengl.util.InputStatus;

/**
 * An entity is an object that can be added to a scene.
 * You can transform its location, rotation and scale.
 * That can have a transformation parent.
 * That can be extended to create new transformable behaviours.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public final class Entity extends Behaviour {
    
    /**
     * The parent scene object.
     */
    private Scene scene;
    
    /**
     * The transformation object that contains the position, 
     * rotation and scale of this object.
     */
    private Transform transform;
    
    /**
     * The list of component to call.
     */
    private List<Component> components;
   
    /**
     * The layout of this entity.
     */
    private String layout;
    
    /**
     * If this entity has been modified
     */
    private boolean modified;

    /**
     * Create an entity with a new transform.
     */
    public Entity() {
        this.transform = new Transform();
        this.components = new ArrayList();
    }
    
    /**
     * Change the parent of this entity.
     * @param parent 
     */
    public final void setParent(final Entity parent) {
        transform.setParent(parent.getTransform());
    }

    public final void setLayout(final String layout) {
        this.layout = layout;
    }

    public final String getLayout() {
        return layout;
    }
    
    public final void setScene(final Scene scene) {
        this.scene = scene;
    }

    public final Scene getScene() {
        return scene;
    }
    
    public final void addComponent(final Component component) {
        component.setParent(this);
        components.add(component);
    }
    
    public final boolean hasComponent(final Class<? extends Component> type) {
        for (Component component : components) {
            if (type.isInstance(component)) {
                return true;
            }
        }
        
        return false;
    }
    
    public final <T extends Component> T getComponent(final Class<T> type) {
        for (Component component : components) {
            if (type.isInstance(component)) {
                return (T) component;
            }
        }
        
        return null;
    }
    
    public final boolean isModified() {
        return modified || transform.isModified();
    }
    
    public final void setModified(final boolean modified) {
        this.modified = modified;
    }

    /**
     * Get the transformation object linked with this entity.
     * @return 
     */
    public final Transform getTransform() {
        return transform;
    }
    
    public final void sendMessage(final String name, final Object object) {
        components.forEach((component) -> {
            component.receiveMessage(name, object);
        });
    }

    @Override
    public void onInit(GL3bc gl) {
        components.forEach((component) -> {
            component.init(gl);
        });
    }

    @Override
    public void onInputting(InputEvent event, InputStatus status) {
        components.forEach((component) -> {
            component.inputting(event, status);
        });
    }

    @Override
    public void onUpdate(GL3bc gl, float delta) {
        components.forEach((component) -> {
            component.update(gl, delta);
        });
    }

    @Override
    public final void onDisplay(final GL3bc gl) {
        components.forEach((component) -> {
            component.display(gl);
        });
    }

    @Override
    public void onRescale(GL3bc gl, int width, int height) {
        components.forEach((component) -> {
            component.rescale(gl, width, height);
        });
    }

    @Override
    public void onDestroy(GL3bc gl) {
        components.forEach((component) -> {
            component.destroy(gl);
        });
    }
    
}
