package fr.amap.lidar.amapvox.visualization.opengl.resources;

import com.jogamp.opengl.GL3bc;

/**
 * Class to create bindable resources.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public abstract class Resource {
    
    /**
     * If this resource has been loaded or not.
     */
    private boolean loaded;
    
    /**
     * The resource identifier used to get it in the resource manager.
     * Each resource has a particular identifier, that can be unique or multiple.
     */
    private final String identifier;

    /**
     * Create a resource with an identifier.
     * @param identifier 
     */
    public Resource(final String identifier) {
        this.identifier = identifier;
    }

    /**
     * Get the linked resourceID.
     * @return 
     */
    public String getIdentifier() {
        return identifier;
    }
    
    /**
     * Call-it to load this object (only if that has not been already loaded).
     * @param gl The current OpenGL interface
     */
    public final void load(final GL3bc gl) {
        if (!loaded) {
            loaded = true;
            onLoad(gl);
        }
    }
    
    /**
     * Implements-it to specify what appends when this object is loaded.
     * @param gl The current OpenGL interface
     */
    protected abstract void onLoad(final GL3bc gl);
    
    /**
     * Call-it to unload this object (work only is that has already been loaded).
     * @param gl The current OpenGL interface
     */
    public final void unload(final GL3bc gl) {
        if (loaded) {
            loaded = false;
            onUnload(gl);
        }
    }
    
    /**
     * Implements-it to specify what appends when this object is unloaded.
     * @param gl The current OpenGL interface
     */
    protected abstract void onUnload(final GL3bc gl);
    
    /**
     * Call-it to bind (update) this object.
     * @param gl The current OpenGL interface
     */
    public final void bind(final GL3bc gl) {
        if (!loaded) {
            load(gl);
        }
        
        onBind(gl);
    }
    
    /**
     * Implements-it to specify what appends when this object is binded (updated).
     * @param gl The current OpenGL interface
     */
    protected abstract void onBind(final GL3bc gl);
    
    /**
     * Call it to unbind this object.
     * @param gl The current OpenGL interface
     */
    public final void unbind(final GL3bc gl) {
        if (loaded) {
            onUnbind(gl);
        }
    }
    
    /**
     * Implements it to specify what appends when this object is unbinded.
     * @param gl The current OpenGL interface
     */
    protected abstract void onUnbind(final GL3bc gl);
    
}
