/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.visualization.opengl.resources.meshes;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import fr.amap.lidar.amapvox.visualization.opengl.resources.Resource;

/**
 *
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public final class Mesh extends Resource {
        
    private Geometry geometry;
    
    /**
     * Which shape to draw the primitive.
     */
    private DrawingType drawingType;
    
    /**
     * How many primitives to draw.
     */
    private int instances;
    
    /**
     * If the back faces are hiden or not (by default true).
     */
    private boolean faceCulling;
    
    /**
     * The depth function used with the depth buffer.
     */
    private DepthFunction depthFunction;
    
    /**
     * If the depth mask is used or not (by default true).
     */
    private boolean depthMask;
    
    private final IntBuffer vao;
    private final IntBuffer vbo;
    
    public Mesh(final String identifier) {
        super(identifier);
        
        this.geometry = null;
        this.drawingType = DrawingType.TRIANGLES;
        this.instances = 1;
        this.faceCulling = true;
        this.depthMask = true;
        this.depthFunction = DepthFunction.LESS;
        
        this.vao = Buffers.newDirectIntBuffer(1);
        this.vbo = Buffers.newDirectIntBuffer(3);
    }

    public final void setGeometry(final Geometry geometry) {
        this.geometry = geometry;
    }

    public final void setDrawingType(final DrawingType drawingType) {
        this.drawingType = drawingType;
    }
    
    /**
     * Specify how many primitives to draw.
     * @param instances 
     */
    public final void setInstances(int instances) {
        this.instances = instances;
    }

    /**
     * If the back faces are hiden or not (by default true).
     * @param faceCulling 
     */
    public final void setFaceCulling(final boolean faceCulling) {
        this.faceCulling = faceCulling;
    }

    /**
     * Set the depth function used with the depth buffer.
     * @param depthFunction 
     */
    public final void setDepthFunction(final DepthFunction depthFunction) {
        this.depthFunction = depthFunction;
    }

    /**
     * If the depth mask is used or not (by default true)
     * @param depthMask 
     */
    public final void setDepthMask(final boolean depthMask) {
        this.depthMask = depthMask;
    }
    
    @Override
    protected void onLoad(final GL3bc gl) {
        if (geometry != null) {
            /* Create vertex arrays */
            gl.glGenVertexArrays(vao.limit(), vao);
            gl.glBindVertexArray(vao.get(0));
            
            /* Create vertex buffers */
            gl.glGenBuffers(vbo.limit(), vbo);
            
            /* Create vertex position buffers */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
            final FloatBuffer vertexBuffer = 
                    Buffers.newDirectFloatBuffer(geometry.getVertices());
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
                    vertexBuffer.limit() * Float.BYTES,
                    vertexBuffer, GL3bc.GL_STATIC_DRAW);
            
            /* Create vertex normal buffers */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
            final FloatBuffer normalBuffer = 
                    Buffers.newDirectFloatBuffer(geometry.getNormals());
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
                    normalBuffer.limit() * Float.BYTES,
                    normalBuffer, GL3bc.GL_STATIC_DRAW);
            
            /* Create vertex normal buffers */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(2));
            final FloatBuffer texcoordBuffer = 
                    Buffers.newDirectFloatBuffer(geometry.getTexcoords());
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
                    texcoordBuffer.limit() * Float.BYTES,
                    texcoordBuffer, GL3bc.GL_STATIC_DRAW);
        }
    }
    
    @Override
    protected void onBind(final GL3bc gl) {
        if (geometry != null) {            
            /* Bind buffer and vertex array */
            gl.glBindVertexArray(vao.get(0));

            /* Enable vertex position attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
            gl.glEnableVertexAttribArray(0);
            gl.glVertexAttribPointer(0, 3, GL3bc.GL_FLOAT, false, 0, 0);

            /* Enable vertex normal attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
            gl.glEnableVertexAttribArray(1);
            gl.glVertexAttribPointer(1, 3, GL3bc.GL_FLOAT, false, 0, 0);

            /* Enable vertex texcoord attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(2));
            gl.glEnableVertexAttribArray(2);
            gl.glVertexAttribPointer(2, 2, GL3bc.GL_FLOAT, false, 0, 0);
        }
    }
    
    /**
     * Draw this mesh by follow these steps :
     * <ol>
     *     <li>Bind the geometry.</li>
     *     <li>Update the face culling.</li>
     *     <li>Enable the depth test.</li>
     *     <li>Update the face depth mask.</li>
     *     <li>Draw each geometry instance with the current rendering method.</li>
     *     <li>Unbind the geometry.</li>
     * </ol>
     * @param gl The current OpenGL interface
     */
    public final void draw(final GL3bc gl) {
        if (geometry != null) {
            /* Bind the geometry */
            bind(gl);
            
            /* Hide back faces if face culling is enabled */
            if (faceCulling) {
                gl.glCullFace(GL3bc.GL_BACK);
                gl.glEnable(GL3bc.GL_CULL_FACE);
            } else {
                gl.glDisable(GL3bc.GL_CULL_FACE);
            }
            
            /* Change the depth function */
            switch(depthFunction) {
                case LEQUAL: 
                    gl.glDepthFunc(GL3bc.GL_LEQUAL);
                    break;
                case LESS:
                    gl.glDepthFunc(GL3bc.GL_LESS);
                    break;
            }
            
            /* Change depth mask */
            gl.glDepthMask(depthMask);
            
            /* Check if the drawing mode is not null */
            if (null != drawingType) {
                /* Get the vertex indices */
                final IntBuffer indexBuffer = Buffers.newDirectIntBuffer(geometry.getIndices());
                
                /* Draw the available and selected drawing mode */
                switch (drawingType) {
                    case POINTS:
                        gl.glDrawArraysInstanced(GL3bc.GL_POINTS, 0, indexBuffer.limit(), instances);
                        break;
                    case LINES:
                        gl.glDrawArraysInstanced(GL3bc.GL_LINES, 0, indexBuffer.limit(), instances);
                        break;
                    case TRIANGLES:
                        gl.glDrawArraysInstanced(GL3bc.GL_TRIANGLES, 0, indexBuffer.limit(), instances);
                        break;
                }
            }
            
            /* Unbind the geometry */
            unbind(gl);
        }
    }
    
    @Override
    protected void onUnbind(final GL3bc gl) {
        /* Unbind buffer and vertex array */
        if (geometry != null) {
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, 0);
            gl.glBindVertexArray(0);
        }
    }

    @Override
    protected void onUnload(final GL3bc gl) {
        /* Clean up */
        if (geometry != null) {
            gl.glDeleteBuffers(vbo.capacity(), vbo);
            gl.glDeleteVertexArrays(vao.capacity(), vao);
        }
    }
    
}
