/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.visualization.opengl.resources.programs;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;
import org.joml.Matrix4f;
import org.joml.Quaternionf;
import org.joml.Vector3f;
import fr.amap.lidar.amapvox.visualization.opengl.resources.Resource;
import fr.amap.lidar.amapvox.visualization.opengl.resources.materials.Material;
import fr.amap.lidar.amapvox.visualization.opengl.resources.meshes.Mesh;
import fr.amap.lidar.amapvox.visualization.opengl.resources.shaders.Shader;

/**
 *
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public abstract class Program extends Resource {
    
    private String layout;
    protected Stack<Material> materials;
    private Mesh mesh;
    protected Shader shader;
    
    final List<Vector3f> positions;
    final List<Quaternionf> rotations;
    final Matrix4f view;
    final Matrix4f projection;
    final Vector3f viewPos;
    float scale;
    
    private int viewLocation;
    private int projLocation;
    private int viewPosLocation;
    private int scaleLocation;
    private IntBuffer vbo;
    
    public Program(final String identifier) {
        super(identifier);
        this.materials = new Stack();
        this.mesh = null;
        this.positions = new ArrayList();
        this.rotations = new ArrayList();
        this.view = new Matrix4f().identity();
        this.projection = new Matrix4f().identity();
        this.viewPos = new Vector3f();
        this.vbo = Buffers.newDirectIntBuffer(2);
        this.scale = 1;
    }
    
    public final void addMaterial(final Material material) {
        this.materials.push(material);
    }

    public final void setMaterials(final List<Material> materials) {
        this.materials.addAll(materials);
    }

    public final void setLayout(final String layout) {
        this.layout = layout;
    }

    public final String getLayout() {
        return layout;
    }
    
    public final void clearMaterials() {
        this.materials.clear();
    }

    public final void setMesh(final Mesh mesh) {
        this.mesh = mesh;
    }

    public final void setScale(final float scale) {
        this.scale = scale;
    }
    
    @Override
    protected void onLoad(final GL3bc gl) {
        mesh.load(gl);
        shader.load(gl);
        shader.bind(gl);
        
        /* Generate the vertex buffer objects */
        gl.glGenBuffers(vbo.limit(), vbo);
        
        /* Generate the position buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
        gl.glEnableVertexAttribArray(3);
        gl.glVertexAttribPointer(3, 3, GL3bc.GL_FLOAT, false, 0, 0);
        gl.glVertexAttribDivisor(3, 1);
        
        /* Generate the roation buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
        gl.glEnableVertexAttribArray(4);
        gl.glVertexAttribPointer(4, 4, GL3bc.GL_FLOAT, false, 0, 0);
        gl.glVertexAttribDivisor(4, 1);
        
        viewPosLocation = gl.glGetUniformLocation(shader.getProgram(), "viewPos");
        viewLocation = gl.glGetUniformLocation(shader.getProgram(), "view");
        projLocation = gl.glGetUniformLocation(shader.getProgram(), "proj");
        scaleLocation = gl.glGetUniformLocation(shader.getProgram(), "scale");
    }
    
    public void updateProperties() {
        /* Get materials values */
        positions.clear();
        rotations.clear();
                
        /* Get each position vector and rotation quaternion */        
        materials.forEach((material) -> {
            positions.add(material.getPosition());
            rotations.add(material.getRotation());
            view.set(material.getView());
            projection.set(material.getProjection());
            viewPos.set(material.getViewPosition());
        });
        
        /* Get the view matrix, the projection matrix 
        and the view position vector */
    }

    @Override
    protected void onBind(final GL3bc gl) {       
        /* Bind the mesh and the shader */
        mesh.bind(gl);
        shader.bind(gl);

        /* Bind the position buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
        final float[] positionsValues = new float[positions.size() * 3];
        for (int i = 0; i < positions.size(); i++) {
            positionsValues[i * 3] = positions.get(i).x;
            positionsValues[i * 3 + 1] = positions.get(i).y;
            positionsValues[i * 3 + 2] = positions.get(i).z;
        }
        final FloatBuffer positionBuf = 
                Buffers.newDirectFloatBuffer(positionsValues);
        gl.glBufferData(GL3bc.GL_ARRAY_BUFFER, Float.BYTES * positionBuf.limit(), 
                positionBuf, GL3bc.GL_DYNAMIC_DRAW);
        
        /* Bind the rotation buffer */
        gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
        final float[] rotationsValues = new float[rotations.size() * 4];
        for (int i = 0; i < rotations.size(); i++) {
            rotationsValues[i * 4] = rotations.get(i).x;
            rotationsValues[i * 4 + 1] = rotations.get(i).y;
            rotationsValues[i * 4 + 2] = rotations.get(i).z;
            rotationsValues[i * 4 + 3] = rotations.get(i).w;
        }
        final FloatBuffer rotationBuf = 
                Buffers.newDirectFloatBuffer(rotationsValues);
        gl.glBufferData(GL3bc.GL_ARRAY_BUFFER, Float.BYTES * rotationBuf.limit(), 
                rotationBuf, GL3bc.GL_DYNAMIC_DRAW);
        
        /* Send the view tranformation matrix */
        final FloatBuffer viewBuf = Buffers.newDirectFloatBuffer(16);
        gl.glUniformMatrix4fv(viewLocation, 1, false, view.get(viewBuf));
        
        /* Send the view tranformation matrix */
        final FloatBuffer projBuf = Buffers.newDirectFloatBuffer(16);
        gl.glUniformMatrix4fv(projLocation, 1, false, projection.get(projBuf));
        
        /* Send the view position vector */
        gl.glUniform3f(viewPosLocation, viewPos.x, viewPos.y, viewPos.z);
        
        /* Send the scale value */
        gl.glUniform1f(scaleLocation, scale);
    }
    
    public final void draw(final GL3bc gl) {
        mesh.setInstances(materials.size());
        mesh.draw(gl);
    }

    @Override
    protected void onUnbind(final GL3bc gl) {
        shader.unbind(gl);
    }

    @Override
    protected void onUnload(final GL3bc gl) {
        shader.unload(gl);
    }
    
}
