/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.visualization.opengl.resources.shaders;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;
import fr.amap.lidar.amapvox.visualization.opengl.resources.Resource;

/**
 *
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public final class Shader extends Resource {

    private static int currentProgram;
    
    /**
     * The program it attached with this shader.
     */
    private int program;

    /**
     * The vertex shader id attached with this object.
     */
    private int vertexShader;

    /**
     * The fragment shader id attached with this object.
     */
    private int fragmentShader;

    /**
     * The vertex GLSL source code.
     */
    private String vertexSources;

    /**
     * The fragment GLSL source code.
     */
    private String fragmentSources;

    public Shader(final String identifier) {
        super(identifier);
        this.program = 0;
        this.vertexShader = 0;
        this.fragmentShader = 0;
        this.vertexSources = "";
        this.fragmentSources = "";
    }

    /**
     * Change the vertex sources.
     * @param paths
     */
    public final void setVertexSources(final String... paths) {
        vertexSources = "";
        
        for (String path : paths) {
            vertexSources += loadGLSLSources(path) + "\n";
        }
    }

    /**
     * Change the fragment sources.
     * @param paths
     */
    public final void setFragmentSources(final String... paths) {
        fragmentSources = "";
        
        for (String path : paths) {
            fragmentSources += loadGLSLSources(path) + "\n";
        }
    }
    
    /**
     * Import the sources of a GLSL file.
     * @param srcPath
     * @return
     */
    private String loadGLSLSources(final String srcPath) {
        try {
            /* Initialize the result */
            String sources = "";
            
            /* Load sources */
            final InputStream sis = getClass().getResourceAsStream(srcPath);
            final InputStreamReader sisr = new InputStreamReader(sis);
            final BufferedReader sbr = new BufferedReader(sisr);
            
            String sl;
            while ((sl = sbr.readLine()) != null) {
                sources += sl + "\n";
            }
            
            /* Return lines */
            return sources;
        } catch (IOException ex) {
            Logger.getLogger(Shader.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        throw(new NullPointerException("The GLSL shader located at : " 
                + srcPath + " is unfoundable"));
    }
    
    /**
     * Get the opengl program id.
     * @return 
     */
    public final int getProgram() {
        return program;
    }
    
    /**
     * Initialize an opengl shader with the specified type and sources.
     * @param type The type of shader to init.
     * @param sources The GLSL source code.
     * @return The opengl shader id.
     */
    private int initShader(final GL3bc gl, final int type, final String sources) {        
        /* Create, attach source to, and compile shader */
        final int shader = gl.glCreateShader(type);
        final String[] lines = sources.split("(?=\n)");
        gl.glShaderSource(shader, lines.length, lines, null, 0);
        gl.glCompileShader(shader);
        
        /* Get the compilation status */
        final IntBuffer compilationStatus = Buffers.newDirectIntBuffer(1);
        gl.glGetShaderiv(shader, GL3bc.GL_COMPILE_STATUS, compilationStatus);

        /* If the compilation has failed, print error logs */
        if (compilationStatus.get() != GL3bc.GL_TRUE) {
            /* Get the info log length */
            final IntBuffer logLength = Buffers.newDirectIntBuffer(1);
            gl.glGetShaderiv(shader, GL3bc.GL_INFO_LOG_LENGTH, logLength);

            /* Get the log, and print-it */
            ByteBuffer infoLog = Buffers.newDirectByteBuffer(logLength.get(0));
            gl.glGetShaderInfoLog(shader, logLength.get(0), logLength, infoLog);
            byte[] b = new byte[logLength.get()];
            infoLog.get(b);
            System.err.println("Error compiling shader n�" + shader 
                    + " -> " + new String(b));
            
            for (int i = 0; i < lines.length; i++) {
                System.out.print(i + ") " + lines[i] + "\n");
            }
        }
        
        return shader;
    }

    @Override
    protected void onLoad(final GL3bc gl) {                   
        if (vertexSources != null && fragmentSources != null) {
            vertexShader = initShader(gl, GL3bc.GL_VERTEX_SHADER, vertexSources);
            fragmentShader = initShader(gl, GL3bc.GL_FRAGMENT_SHADER, fragmentSources);
            
            program = gl.glCreateProgram();
            gl.glAttachShader(program, vertexShader);
            gl.glAttachShader(program, fragmentShader);
            gl.glLinkProgram(program);
            
            /* Get the linking status */
            final IntBuffer linkStatus = Buffers.newDirectIntBuffer(1);
            gl.glGetProgramiv(program, GL3bc.GL_LINK_STATUS, linkStatus);
            
            if (linkStatus.get() != GL3bc.GL_TRUE) {
                /* Get the info log length */
                final IntBuffer logLength = Buffers.newDirectIntBuffer(1);
                gl.glGetProgramiv(program, GL3bc.GL_INFO_LOG_LENGTH, logLength);

                /* Get the log, and print-it */
                final ByteBuffer infoLog = Buffers.newDirectByteBuffer(logLength.get(0));
                gl.glGetProgramInfoLog(program, logLength.get(0), logLength, infoLog);
                final byte[] b = new byte[logLength.get()];
                infoLog.get(b);
                System.err.println("Error linking program n�" + program 
                        + " -> " + new String(b));
            }
        }
    }

    @Override
    protected void onBind(final GL3bc gl) {        
        /* Bind the attached program */
        if (currentProgram != program) {
            currentProgram = program;
            gl.glUseProgram(program);
        }
    }

    @Override
    protected void onUnbind(final GL3bc gl) {
        gl.glUseProgram(0);
    }

    @Override
    protected void onUnload(final GL3bc gl) {        
        /* Clean up shaders */
        if (vertexShader > 0) {
            gl.glDeleteShader(vertexShader);
        }
        
        if (fragmentShader > 0) {
            gl.glDeleteShader(fragmentShader);
        }
         /* Clean up program */
        if (program > 0) {
            gl.glDeleteProgram(program);
        }
    }

}
