/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.visualization.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Properties;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 *
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class ProjectProperties {

    private static Properties properties;

    private static final String COMMENTS
            = "AMAPVox Visualization properties file (generated automatically)";
    private static final String DIRECTORY_NAME
            = System.getProperty("user.home") + "/.amapvox/";
    private static final String FILE_NAME
            = "project.properties";

    /**
     * Set a property value in the project properties file. This will save the
     * file.
     * @param key
     * @param value
     */
    public static final synchronized void setProperty(final String key,
            final String value) {
        if (properties == null) {
            loadFile();
        }

        properties.setProperty(key, value);
        saveFile();
    }

    /**
     * Get a property from the properties file. This will generate a file if
     * that is unfound and add the default value of its property's key.
     * @param key
     * @param defaultValue
     * @return
     */
    public static final synchronized String getProperty(final String key, 
            final String defaultValue) {
        if (properties == null) {
            loadFile();
        }

        if (properties.containsKey(key)) {
            return properties.getProperty(key, defaultValue);
        } else {
            setProperty(key, defaultValue);
            return defaultValue;
        }
    }

    /**
     * Load the properties file if that exist or create it by calling the
     * saveFile method.
     */
    private static synchronized void loadFile() {
        properties = new Properties();

        final File directory = new File(DIRECTORY_NAME);
        if (!directory.exists()) {
            directory.mkdirs();
        }

        final File file = new File(DIRECTORY_NAME + FILE_NAME);
        if (file.exists()) {
            final ExecutorService executor
                    = Executors.newCachedThreadPool();

            executor.execute(() -> {
                try {
                    final InputStream inputStream
                            = new FileInputStream(file);
                    properties.load(inputStream);
                } catch (IOException e) {
                    /* Ignore this error */
                }
            });

            executor.shutdown();
        } else {
            saveFile();
        }
    }

    /**
     * Save the properties file at the specified directory. This will replace
     * existing file.
     */
    private static synchronized void saveFile() {
        final ExecutorService executor
                = Executors.newCachedThreadPool();
        
        executor.execute(() -> {
            
            final File directory = new File(DIRECTORY_NAME);
            if (!directory.exists()) {
                directory.mkdirs();
            }

            final File file = new File(DIRECTORY_NAME + FILE_NAME);
            if (file.exists()) {
                file.delete();
            }

            try {
                file.createNewFile();
                final OutputStream outputStream = new FileOutputStream(file);
                properties.store(outputStream, COMMENTS);
            } catch (IOException e) {
                e.printStackTrace();
            }
        });

        executor.shutdown();
    }

}
