/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.amap.lidar.amapvox.visualization.views;

import fr.amap.commons.util.debug.DebugLevel;
import fr.amap.commons.util.debug.Debugger;
import fr.amap.lidar.amapvox.visualization.util.View;
import java.awt.BorderLayout;
import java.awt.ComponentOrientation;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import javax.swing.Action;
import javax.swing.JCheckBox;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.ScrollPaneConstants;
import org.jdesktop.swingx.JXButton;
import org.jdesktop.swingx.JXCollapsiblePane;
import org.jdesktop.swingx.JXComboBox;
import org.jdesktop.swingx.JXLabel;
import org.jdesktop.swingx.JXPanel;
import org.jdesktop.swingx.JXTitledPanel;
import fr.amap.lidar.amapvox.visualization.controllers.RenderingController;
import fr.amap.lidar.amapvox.visualization.opengl.util.ColorBars;
import fr.amap.lidar.amapvox.visualization.opengl.util.Skyboxes;
import fr.amap.lidar.amapvox.visualization.util.ProjectProperties;

/**
 *
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class RenderingView extends View {

    private final RenderingController controller;

    private JXCollapsiblePane collapsiblePane;
    private JXTitledPanel titledPanel;
    private JXPanel titledPanelContent;
    private JScrollPane titledPanelScrollPane;

    private JXPanel backgroundPanel;
    private JXLabel backgroundLabel;
    private JXComboBox backgroundComboBox;
    
    private JXPanel reflectionPanel;
    private JCheckBox reflectionCheckbox;
    
    private JXPanel colorbarPanel;
    private JXLabel colorbarLabel;
    private JXComboBox colorbarComboBox;
    
    private JXPanel drawFastPanel;
    private JCheckBox drawFastCheckbox;
    
    private JXPanel voxelSizePanel;
    private JXLabel voxelSizeLabel;
    private JSlider voxelSizeSlider;
    
    /**
     * Creates new form RenderingFrame
     * @throws java.lang.Exception
     */
    public RenderingView() throws Exception {
        super();
        this.controller = new RenderingController();
    }

    public final RenderingController getController() {
        return controller;
    }
    
    public final void updateInitialProperties() {
        controller.changeEnableReflection(reflectionCheckbox.isSelected());
        controller.changeVoxelSize(voxelSizeSlider.getValue());
    }

    @Override
    protected void initContainer() throws Exception {
        setLocationRelativeTo(getParent());
    }

    @Override
    protected void initComponents() throws Exception {
        /* Create the collapsible panel */
        collapsiblePane = new JXCollapsiblePane(JXCollapsiblePane.Direction.LEFT);
        collapsiblePane.setAnimated(false);
        collapsiblePane.setLayout(new BorderLayout());

        final Action toggleAction = collapsiblePane.getActionMap().
                get(JXCollapsiblePane.TOGGLE_ACTION);

        toggleAction.putValue(JXCollapsiblePane.COLLAPSE_ICON,
                new javax.swing.ImageIcon(getClass().getResource("/icons/right.png")));
        toggleAction.putValue(JXCollapsiblePane.EXPAND_ICON,
                new javax.swing.ImageIcon(getClass().getResource("/icons/left.png")));

        final JXButton toggle = new JXButton(toggleAction);
        toggle.setText(null);
        toggle.setPreferredSize(new Dimension(20, 20));

        add(collapsiblePane, BorderLayout.CENTER);
        add(toggle, BorderLayout.WEST);

        /* Create title panel */
        titledPanel = new JXTitledPanel();
        titledPanel.getContentContainer().setLayout(new FlowLayout());
        addLocaleCallback((locale) -> {
            titledPanel.setTitle(bundle.getString("renderingView_Title"));
        });
        collapsiblePane.getContentPane().add(titledPanel);

        /* Create scroll pane */
        titledPanelContent = new JXPanel();
        titledPanelContent.setLayout(new GridLayout(0, 1));

        titledPanelScrollPane = new JScrollPane();
        titledPanelScrollPane.setHorizontalScrollBarPolicy(
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        titledPanelScrollPane.setViewportView(titledPanelContent);
        titledPanel.getContentContainer().add(titledPanelScrollPane);

        /* Create background components */
        backgroundPanel = new JXPanel(new FlowLayout(FlowLayout.LEFT));

        backgroundLabel = new JXLabel();
        addLocaleCallback((locale) -> {
            backgroundLabel.setText(bundle.getString("renderingView_backgroundLabel") + " :");
        });

        backgroundComboBox = new JXComboBox(Skyboxes.getSkyNames());
        backgroundComboBox.setSelectedItem(ProjectProperties.getProperty("background", "Simple"));
        backgroundComboBox.addActionListener((java.awt.event.ActionEvent evt) -> {
            try {
                /* Change the background of the scene */
                controller.changeBackground((String) backgroundComboBox.getSelectedItem());
            } catch (IOException ex) {
                Debugger.debugInterface(ex, DebugLevel.ERROR, parentView);
            }
        });

        backgroundPanel.add(backgroundLabel);
        backgroundPanel.add(backgroundComboBox);
        titledPanelContent.add(backgroundPanel);
        
        /* Create reflection components */
        reflectionPanel = new JXPanel(new FlowLayout(FlowLayout.LEFT));
        titledPanelContent.add(reflectionPanel);
        
        reflectionCheckbox = new JCheckBox();
        reflectionCheckbox.setComponentOrientation(ComponentOrientation.RIGHT_TO_LEFT);
        addLocaleCallback((locale) -> {
            reflectionCheckbox.setText(bundle.getString("renderingView_reflectionCheckbox"));
        });
        reflectionCheckbox.addActionListener((e) -> {
            controller.changeEnableReflection(reflectionCheckbox.isSelected());
        });
        reflectionCheckbox.setSelected(Boolean.valueOf(
                ProjectProperties.getProperty("reflectionsEnabled", "true")));
        reflectionPanel.add(reflectionCheckbox);

        /* Create colorbar components */
        colorbarPanel = new JXPanel(new FlowLayout(FlowLayout.LEFT));

        colorbarLabel = new JXLabel();
        addLocaleCallback((locale) -> {
            colorbarLabel.setText(bundle.getString("renderingView_colorbarLabel") + " :");
        });

        colorbarComboBox = new JXComboBox(ColorBars.getColorBarsNames());
        colorbarComboBox.setSelectedItem(ProjectProperties.getProperty("colorBar", "Equatorial"));
        colorbarComboBox.addActionListener((e) -> {
            /* Change the color bar of the voxel space */
            controller.changeColorBar((String) colorbarComboBox.getSelectedItem());
        });

        colorbarPanel.add(colorbarLabel);
        colorbarPanel.add(colorbarComboBox);
        titledPanelContent.add(colorbarPanel);
        
        /* Create draw fast components */
        drawFastPanel = new JXPanel(new FlowLayout(FlowLayout.LEFT));
        
        drawFastCheckbox = new JCheckBox();
        drawFastCheckbox.setComponentOrientation(ComponentOrientation.RIGHT_TO_LEFT);
        drawFastCheckbox.setSelected(Boolean.valueOf(
                ProjectProperties.getProperty("drawFastEnabled", "true")));
        addLocaleCallback((locale) -> {
            drawFastCheckbox.setText(bundle.getString("renderingView_drawFastCheckbox"));
        });
        drawFastCheckbox.addActionListener((e) -> {
            controller.changeDrawFast(drawFastCheckbox.isSelected());
        });
        
        drawFastPanel.add(drawFastCheckbox);
        titledPanelContent.add(drawFastPanel);
        
        /* Create voxel size components */
        voxelSizePanel = new JXPanel(new FlowLayout(FlowLayout.LEFT));
        
        voxelSizeLabel = new JXLabel();
        addLocaleCallback((locale) -> {
            voxelSizeLabel.setText(bundle.getString("renderingView_voxelSizeSlider") + " :");
        });
        
        voxelSizeSlider = new JSlider(0, 100);
        voxelSizeSlider.setValue(Integer.valueOf(
                ProjectProperties.getProperty("voxelSize", "100")));
        voxelSizeSlider.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseReleased(MouseEvent e) {
                controller.changeVoxelSize(voxelSizeSlider.getValue());
            }
        });
        
        voxelSizePanel.add(voxelSizeLabel);
        voxelSizePanel.add(voxelSizeSlider);
        titledPanelContent.add(voxelSizePanel);

        pack();
    }

}
