 /*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.visualization.views;

import fr.amap.commons.util.swing.jprogress.JProgressEvent;
import fr.amap.commons.util.swing.jprogress.JProgressFrame;
import java.awt.Color;
import java.io.IOException;
import java.io.InputStream;
import org.joml.Vector4f;
import fr.amap.lidar.amapvox.visualization.opengl.entities.base.Entity;
import fr.amap.lidar.amapvox.visualization.opengl.entities.base.Scene;
import fr.amap.lidar.amapvox.visualization.opengl.entities.extended.Camera;
import fr.amap.lidar.amapvox.visualization.opengl.entities.extended.CameraController;
import fr.amap.lidar.amapvox.visualization.opengl.entities.extended.MeshRenderer;
import fr.amap.lidar.amapvox.visualization.opengl.entities.extended.VoxelSpace;
import fr.amap.lidar.amapvox.visualization.opengl.entities.util.VoxelSpaceFile;
import fr.amap.lidar.amapvox.visualization.opengl.resources.materials.StandardMaterial;
import fr.amap.lidar.amapvox.visualization.opengl.resources.meshes.Geometry;
import fr.amap.lidar.amapvox.visualization.opengl.resources.meshes.Mesh;
import fr.amap.lidar.amapvox.visualization.opengl.resources.postprocesses.extended.GammaCorrectionPostProcess;
import fr.amap.lidar.amapvox.visualization.opengl.resources.postprocesses.extended.SharpenPostProcess;
import fr.amap.lidar.amapvox.visualization.opengl.resources.textures.Texture2D;
import fr.amap.lidar.amapvox.visualization.opengl.resources.textures.TextureFormat;
import fr.amap.lidar.amapvox.visualization.opengl.util.ColorBars;
import fr.amap.lidar.amapvox.visualization.opengl.util.Skyboxes;
import fr.amap.lidar.amapvox.visualization.util.ProjectProperties;

/**
 *
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
public class VisualizationScene extends Scene {
    
    /* Resources */
    private Mesh cubeMesh;
    private Mesh axisMesh;
    private Texture2D axisAlbedo;
    private Texture2D voxelAlbedo;
    
    /* Entities */
    private Entity cameraControllerEntity;
    private CameraController cameraController;
    private Entity mainCameraEntity;
    private Camera mainCamera;
    private Entity miniCameraEntity;
    private Camera miniCamera;
    private Entity axisEntity;
    private MeshRenderer axisRenderer;
    private Entity voxelSpaceEntity;
    private VoxelSpace voxelSpace;
    
    private Runnable changeFileCallback;
    
    public VisualizationScene() throws IOException {
        super();
        initResources();
        initEntities();
    }
    
    public final void setVoxelSpaceFile(final InputStream in, final JProgressFrame progressFrame) {
        if (voxelSpace != null) {
            final VoxelSpaceFile file = new VoxelSpaceFile();
            
            if (progressFrame != null) {
                file.addProgressListener((JProgressEvent e) -> {
                    progressFrame.setProgression(e.getValue(), e.getMax(), e.getMessage());
                });
            }
            
            file.parse(in);
            voxelSpace.setFile(file);
            mainCameraEntity.getTransform().getPosition().z = 
                voxelSpace.getCameraRetreat(60);
            somethingsModified = true;
            
            if (changeFileCallback != null) {
                changeFileCallback.run();
            }
        }
    }

    public void setChangeFileCallback(final Runnable changeFileCallback) {
        this.changeFileCallback = changeFileCallback;
    }
    
    public final VoxelSpace getVoxelSpaceComponent() {
        return voxelSpace;
    }
    
    private void initResources() throws IOException {
        cubeMesh = new Mesh("CubeMesh");
        cubeMesh.setGeometry(Geometry.loadFromOBJ("/meshes/cube.obj"));

        axisMesh = new Mesh("AxisMesh");
        axisMesh.setGeometry(Geometry.loadFromOBJ("/meshes/axis.obj"));

        axisAlbedo = new Texture2D("AxisAlbedo", 256, 256,
                TextureFormat.RGBA, true, "/textures/axisAlbedo.png");

        voxelAlbedo = new Texture2D("VoxelAlbedo", 256, 256,
                TextureFormat.RGBA, true, "/textures/voxelAlbedo.png");

        setBackground(Skyboxes.getSky(ProjectProperties.getProperty("background", "Simple")));
    }
    
    private void initEntities() throws IOException {
        /* Create the voxel space */
        voxelSpaceEntity = new Entity();
        voxelSpaceEntity.setLayout("voxel");
        voxelSpace = new VoxelSpace(cubeMesh);
        voxelSpace.setAlbedo(voxelAlbedo);
        voxelSpace.setColorBar(ColorBars.getColorBar(
                ProjectProperties.getProperty("colorBar", "Equatorial")));
        voxelSpace.setDrawFastEnabled(Boolean.valueOf(
                ProjectProperties.getProperty("drawFastEnabled", "true")));
        voxelSpaceEntity.addComponent(voxelSpace);
        addEntity(voxelSpaceEntity);
        
        /* Create the camera controller */
        cameraControllerEntity = new Entity();
        cameraController = new CameraController();
        cameraControllerEntity.addComponent(cameraController);
        addEntity(cameraControllerEntity);

        /* Create the main camera */
        mainCameraEntity = new Entity();
        mainCameraEntity.setParent(cameraControllerEntity);
        
        mainCamera = new Camera();
        mainCamera.setZoomEnabled(true);
        mainCamera.addLayout("background");
        mainCamera.addLayout("voxel");
        mainCamera.addPostProcess(0, new SharpenPostProcess());
//        mainCamera.addPostProcess(1, new TonemappingPostProcess());
        //mainCamera.addPostProcess(2, new FXAAPostProcess());
        mainCamera.addPostProcess(3, new GammaCorrectionPostProcess());
        
        mainCameraEntity.addComponent(mainCamera);
        addEntity(mainCameraEntity);
        
        /* Create the axis */
        axisRenderer = new MeshRenderer();
        final StandardMaterial material = new StandardMaterial();
        material.setAlbedo(axisAlbedo);
        material.setMainColor(Color.WHITE);
        axisRenderer.setMesh(axisMesh);
        axisRenderer.setMaterial(material);
        
        axisEntity = new Entity();
        axisEntity.setLayout("axis");
        axisEntity.addComponent(axisRenderer);
        addEntity(axisEntity);
        
        /* Create the mini camera */
        miniCamera = new Camera();
        miniCamera.addLayout("axis");
        miniCamera.setProportions(new Vector4f(0.79f, 0.01f, 0.2f, 0.2f));
        //miniCamera.addPostProcess(0, new SharpenPostProcess());
        //miniCamera.addPostProcess(3, new GammaCorrectionPostProcess());
        
        miniCameraEntity = new Entity();
        miniCameraEntity.setParent(cameraControllerEntity);
        miniCameraEntity.getTransform().getPosition().z = 2.5f;
        miniCameraEntity.addComponent(miniCamera);
        addEntity(miniCameraEntity);
    }

}
