/**
 * Included vertex shader that contains all needed attributes, 
 * uniforms and methods for each vertex shader.
 */

/* Use OpenGL version 3.3 compatibility to get a wide flexibility */
#version 330 compatibility

/* ---------------------------------------------------------------------- */
/* attributes area */
/* ---------------------------------------------------------------------- */
 
/* Vertex position attribute */
layout (location = 0) in vec3 vertex;

/* Vertex normal attribute */
layout (location = 1) in vec3 normal;

/* Vertex texture coordinate attribute */
layout (location = 2) in vec2 texcoord;

/* Transformation position vector */
layout (location = 3) in vec3 position;

/* Transformation rotation quaternion */
layout (location = 4) in vec4 rotation;

/* ---------------------------------------------------------------------- */
/* Uniforms area */
/* ---------------------------------------------------------------------- */

/* View transformation matrix */
uniform mat4 view;

/* Projection matrix */
uniform mat4 proj;

/* Scale float */
uniform float scale;

/* ---------------------------------------------------------------------- */
/* Outputs area */
/* ---------------------------------------------------------------------- */

/* View space position vector */
out vec3 fragPosition;

/* View space normal vector */
out vec3 fragNormal;

/* View space texture coordinate vector */
out vec2 fragTexcoord;

/* ---------------------------------------------------------------------- */
/* Methods area */
/* ---------------------------------------------------------------------- */

/**
 * Transform the position vector to a translation matrix.
 */
mat4 getPositionMatrix()
{
    vec3 p = position;
    return mat4(1,0,0,0,0,1,0,0,0,0,1,0,p.x, p.y, p.z,1);
}

/**
 * Transform the rotation quaternion to rotation matrix.
 */
mat4 getRotationMatrix()
{
    vec4 q = rotation;

    float sqw = q.w*q.w;
    float sqx = q.x*q.x;
    float sqy = q.y*q.y;
    float sqz = q.z*q.z;

    // invs (inverse square length) is only required 
    // if quaternion is not already normalised
    float invs = 1 / (sqx + sqy + sqz + sqw);
    float m00 = ( sqx - sqy - sqz + sqw)*invs;
    float m11 = (-sqx + sqy - sqz + sqw)*invs;
    float m22 = (-sqx - sqy + sqz + sqw)*invs;
    
    float tmp1 = q.x*q.y;
    float tmp2 = q.z*q.w;
    float m10 = 2.0 * (tmp1 + tmp2)*invs;
    float m01 = 2.0 * (tmp1 - tmp2)*invs;
    
    tmp1 = q.x*q.z;
    tmp2 = q.y*q.w;
    float m20 = 2.0 * (tmp1 - tmp2)*invs;
    float m02 = 2.0 * (tmp1 + tmp2)*invs;
    tmp1 = q.y*q.z;
    tmp2 = q.x*q.w;
    float m21 = 2.0 * (tmp1 + tmp2)*invs;
    float m12 = 2.0 * (tmp1 - tmp2)*invs;

    return mat4(
        m00, m01, m02,   0,
        m10, m11, m12,   0,
        m20, m21, m22,   0,
          0,   0,   0,   1
    );
}

/**
 * Get the model matrix by transforming the position vector and 
 * the rotation quaternion to a transformation matrix.
 */
mat4 getModelMatrix()
{
    return getPositionMatrix() * getRotationMatrix();
}